----- Helpers

local function clamp(value, min_value, max_value)
  return max(min(value, max_value), min_value)
end

-- local we have to duplicate
local function length(dx,dy)
  return sqrt(dx*dx+dy*dy)
end

local function length_to(a, b)
  ax, ay = object_center(a)
  bx, by = object_center(b)
  return length(ax-bx, ay-by)
end

-- local we have to duplicate
local function object_point_dist(object,x1,y1)
  local x0,y0=object_center(object)
  return length(x0-x1,y0-y1)
end

-- local we have to duplicate
local function nearest(object_iter,x,y)
  local rval
  local dist
  for obj in object_iter() do
    if(not dist)
      dist=object_point_dist(obj,x,y)
      rval=obj
    else
      local mydist=object_point_dist(obj,x,y)
      if(mydist<dist)
        dist=mydist
        rval=obj
      end
    end
  end

  return rval
end

----- Ship Evaluation

local function is_combat_ship(ship)
  for w in weapons(ship)
    if alive_weapons(w) > 0
      return true
    end
  end
  return false
end

-- returns true if this ship should attempt to board in combat
local function is_combat_boarder(ship)
  local remaining_crew = ship.max_crew - ship.crew_damage
  if remaining_crew < 15
    return false
  end

  -- same check for functional assault shuttles or transporter used in boarding.lua
  local range, sys = ship:boarding_range()
  if range > 0
    return true, sys
  end

  return ship.has_boarding_module
end

-- don't count ships whose weapons are destroyed, as combat ships
function enemy_combat_ships()
  local f = current_enemy_ships ()
  return function(s,v)
    local a = f(s,v)
    while a
      if is_combat_ship(a) return a end
      --print('skipping',a)
      a=f(s,a)
    end
  end
end


-- TODO factor in weapon mods if needed

-- average damage per mount
local function average_weapon_damage(weapon)
  if weapon.is_fighter
    --while hangars bays are not counted for attack value, handling hangars here
    --is important for deciding whether or not to reserve interceptors
    local ship = weapon.object
    local fighter_weapon = get_fighter_weapon(weapon.parent_type, ship.empire, ship)
    return WEAPONS[fighter_weapon].damage_fun.average
  end

  return weapon.info.damage_fun.average
end

-- average damage per subobject
local function average_tracker_damage(tracker)
  if(tracker.tracker_type == 'missile')
    if(tracker.submunitions > 0)
      return WEAPONS[tracker.submunitions_weapon].damage_fun.average * tracker.submunitions
    end
    return WEAPONS[tracker.warhead.parent_type].damage_fun.average
  end
  
  if(tracker.tracker_type == 'fighter')
    local total = 0
    --fighters only ever have 1 weapon right now, but just in case...
    --(yes even strike fighters, their weapon is chosen on launch based on the dest type)
    for i,w in ipairs(tracker.weapons) do
      total += WEAPONS[w.parent_type].damage_fun.average
    end
    return total
  end

  return 0
end

-- a score that represents how "dangerous" a potential target is
function calc_attack_value(ship)
  local score = 0

  for w in weapons(ship) do
    local damage = average_weapon_damage(w)
    if w.is_beam
      damage *= 0.67 --beam attenuation fudge factor
    end
    if w.is_launcher
      damage *= 0.85 --PD fudge factor
    end

    local alive = alive_weapons(w)
    if w.is_fighter
      damage *= 0.5 --PD and flight time fudge factor

      -- hangar bays don't track the number of inflight fightercraft, so just use w.missing
      -- it does mean that carriers with inflight craft have lower priority, but that's better than
      -- carriers who have no more craft retaining a high priority
      alive = max(0, alive - w.missing)
    end
    score += damage * alive
  end

  return score
end


local valuable_system_types = {
  shield_regenerator = true,
  shield = true,
  capacitor = true,
  reactor = true,
  shuttle = true,
  boarding_module = true
}

local function system_value(s)
  if valuable_system_types[s.info.type]
    return 1
  end
  if s.parent_type == 'Transporter' -- Transporter's system type is 'system' for some reason
    return 1
  end
  if MUNITION[s.parent_type]
    return 0.5 --killing this can destroy the ship
  end
  return 0
end

-- a score that represents how "tanky" a potential target is
-- we want this to reflect how much "useless" damage we have to chew through before we start damaging systems
-- so we only include shields and armor
function calc_defence_value(ship)
  -- ignore shield and armour reflection, for simplicity

  local shield_score = 0
  if ship.max_shields > 0
    local shield_regen = floor(ship.adjusted_max_shields*ship.shield_regen_rate)*2

    -- use rms to weight highly damaged shield segments more heavily
    -- that way if a shield segment is down it counts more than a straight average
    local rms_shield_dam = 0
    for i,v in ipairs(SHIELD_NAMES)
      rms_shield_dam += ship[v]*ship[v]
    end
    rms_shield_dam = sqrt(rms_shield_dam / #SHIELD_NAMES)

    shield_score = ship.max_shields - rms_shield_dam + shield_regen
  end

  local armor_score = ship.armor - ship.armor_damage

  local useless_systems = 0
  for s in systems(ship) do
    if s.damage < s.health -- destroyed systems can't soak up damage
      useless_systems += (1-system_value(s))
    end
  end
  
  return shield_score + armor_score + useless_systems
end

function desired_weapon_range(w)
    if w.is_fighter
      return w:launcher_range() -- fightercraft want to be within launch range to minimize time to target
    end
    if w.is_anti_missile
      return ranges.anti_missile
    end
    if w.type=='energy_torpedo'
      return w:launcher_range() * 0.8
    end
    if w.is_launcher
      return w:launcher_range() * 3
    end
    if w.is_pd
      return ranges.short
    end
    if w.is_beam
      return ranges.short/w:effective_distance(1)
    end
    return ranges.medium/w:effective_distance(1)
end

function desired_attack_dist(ship)
  if ship.weapons?[1]?.parent_type=='Suicide Attack'
    return 3
  end

  local totalweight = 0
  local total = 0

  local totalweight_pd = 0
  local total_pd = 0

  for w in weapons(ship) do
    local range = desired_weapon_range(w)

    local damage = average_weapon_damage(w) * alive_weapons(w)
    local weight = damage * damage

    if(w.is_pd)
      total_pd += weight * range
      totalweight_pd += weight
    else
      total += weight * range
      totalweight += weight
    end
  end

  if(totalweight == 0) --this ship only has PD weapons
    total = total_pd
    totalweight = totalweight_pd
  end

  --print("DESIRED ATTACK DISTANCE for", ship.name, "-->", total / totalweight)

  return (totalweight > 0 and total / totalweight) or nil
end


----- Boarding

local function unboardable(ship)
  return ship.drawer_table?.unboardable or ship.is_planet
end


-- copied from boarding_actions.lua
local function crew_combat_strength(ship)
  local rval=0
  rval+= ( ship.empire.tactics.boarding_bonus or 0 )
  return rval
end

local function estimated_boarding_crew_damage(ship)
    local boarding_crew = max(ship.max_crew - ship.crew_damage, 0)
    local crew_hitchance = crew_combat_strength(ship)
    return boarding_crew*(50 + crew_hitchance)/100 * 0.5 * 0.5
end

-- if the target's crew is much larger than ours and
-- they are not disabled we attempt to raid them instead
local function boarding_action(ship, target)
  local boarders = ship.max_crew-ship.crew_damage
  local defenders = target.max_crew-target.crew_damage

  -- if we don't think we can capture the target in 3 boarding actions just raid instead
  local hack = estimated_boarding_crew_damage(ship)*2
  if(is_combat_ship(target) and ceil(defenders/hack) > 3)
    action.raid(ship,target)
  else
    action.capture(ship,target)
  end
end

local function capture_if_able(ship,target)
  --make sure ships don't run out of crew
  local boarders = ship.max_crew-ship.crew_damage
  local defenders = target.max_crew-target.crew_damage
  if boarders < 15 or (boarders < 30 and boarders < defenders)
    return
  end

  if(target.is_planet)
    return
  end

  if(ship.dead or target.dead) 
    return true
  end

  if(ship.has_boarded)
    return false
  end

  if(can_board(ship,target))
    boarding_action(ship,target)
    return not ship.has_boarded
  end

  local x,y,r=find_max_boarding_move(ship,target,-1)

  if(x)

    if(ship.fuel==0)
      return false
    end

    action.move_ships{ship,x,y,r}

    if(ship.dead or target.dead) 
      return true
    end

    if(can_board(ship,target))
      boarding_action(ship,target)
    end
  end

  return not ship.has_boarded
end

-- boarding doesn't care about defence value, only the amount of defending crew
-- in fact, targets with higher defence value are more appealing since destroying
-- them would otherwise be more difficult
function boarding_priority(target)
  if unboardable(target)
    return 0
  end
  local remaining_crew = max(target.max_crew - target.crew_damage, 0)
  return calc_attack_value(target)/(1 + remaining_crew)
end

-- calculate boarding hazard
-- this is used to decide how dangerous combat boarders are.
-- the more heavily armoured and shielded our ships are, the more dangerous boarders
-- are as perceived by the AI, since boarders bypass those things.
--
-- This value is added to the attack value of the ship so we want it to use analogous "units"
-- in this case, the amount of damage that would be bypassed by boarding (the defence value)
--
-- We also multiply by (estimated crew damage/defending crew), but we can't
-- know the estimated crew damage without having the attacking ship
-- so just divide by defending crew here, and multiply by estimated casualties later.
local function calc_boarding_hazard(iter)

  local total = 0
  local ship_count = 0
  for obj in iter() do
    if is_ship(obj)
      if not unboardable(obj)
        local remaining_crew = max(obj.max_crew - obj.crew_damage, 0)
        local boarding_value = calc_defence_value(obj) / remaining_crew
        total += boarding_value*boarding_value --use rms so that the most highly valuable ships are weighted more
      end
      ship_count += 1
    end
  end

  return (ship_count > 0 and sqrt(total/ship_count)) or 0
end



----- Combat AI Methods

local CombatAIMethods = {}

function CombatAIMethods.init_turn()
  local new_turn = setmetatable({}, {__index = CombatAIMethods})

  -- need to get an object belonging to the current side for enemies()
  -- would be nice if enemies() accepted current_side() as an argument...
  for obj in objects() do
    if obj.empire == current_side()
      new_turn.friendly_obj = obj
      break
    end
  end

  new_turn.boarding_hazard = calc_boarding_hazard(current_sides_ships) --how threatening enemy boarders are
  new_turn.boarding_power = calc_boarding_hazard(enemy_combat_ships)   --how valuable our boarders are

  -- lazy-populate target priorities
  new_turn.ship_priority = setmetatable({}, {
      __index = function(t, k)
        new_turn:update_ship_priority(k)
        return t[k]
      end
    })

  -- find the priority of the most valuable friendly ship
  -- used as a hypothetical target for hostile launchers
  -- find the priority of the most valuable/vulnerable friendly ship
  local max_priority = 0
  for ship in current_sides_ships() do
    print(ship.name, ":", new_turn.ship_priority[ship])
    max_priority = max(max_priority, new_turn.ship_priority[ship])
  end
  new_turn.highest_friendly_priority = max_priority

  new_turn:build_target_list()
  new_turn:build_tracker_lists()
  new_turn:build_launcher_lists()

  -- table mapping objects => list of friendly pd/interceptors reserved for that object
  new_turn.reserved_weapons = {}

  return new_turn
end

function CombatAIMethods:is_hostile(obj)
  if(not self.friendly_obj)
    return true
  end
  return enemies(self.friendly_obj, obj)
end

----- Target Lists

function CombatAIMethods:build_target_list()
  self.targets = {}

  -- attack enemy combat ships
  for target in enemy_combat_ships() do
    table.insert(self.targets, target)
  end

  -- if there are no combat ships left, then attack all enemy ships
  if(not next(self.targets))
    for target in current_enemy_ships() do
      table.insert(self.targets, target)
    end
  end
end

function CombatAIMethods:drop_target(target)
  --remove from the target list
  for i,t in ipairs(self.targets)
    if t == target
      table.remove(self.targets, i)
      break
    end
  end

  --also drop any launchers attached to that target
  --need to iterate this in reverse
  for i=#self.hostile_launchers,1,-1 do
    local l = self.hostile_launchers[i]
    if l.object == target
      table.remove(self.hostile_launchers, i)
    end
  end
end

----- Tracker Lists

function CombatAIMethods:build_tracker_lists()

  -- list of hostile trackers
  self.hostile_trackers = {}

  -- table mapping objects => list of friendly trackers targeting that object
  -- see register_enroute_tracker()
  self.enroute_trackers = {}

  -- populate lists
  for obj in objects() do
    if is_tracking_object(obj)
      if self:is_hostile(obj)
        table.insert(self.hostile_trackers, obj)
      end
      if enemies(obj, obj.dest)
        self:register_enroute_tracker(obj, obj.dest)
      end
    end
  end

end

function CombatAIMethods:build_launcher_lists()
  -- list of hostile launchers
  -- used when deciding to reserve interceptors for dangerous trackers not yet launched
  self.hostile_launchers = {}
  self.hostile_shuttles = {} --because assault shuttle bays are not weapons :(

  for ship in current_enemy_ships() do
    for w in weapons(ship)
      if w.is_launcher and alive_weapons(w) > 0
        table.insert(self.hostile_launchers, w)
      end
    end

    --shuttles are not launchers, unfortunately
    local boarder, sys = is_combat_boarder(ship)
    if boarder and sys=='Assault Shuttles'
      table.insert(self.hostile_shuttles, ship)
    end
  end
end

--register a friendly tracker in-flight
function CombatAIMethods:register_enroute_tracker(tracker, target)
  local tracker_data = { damage_per=average_tracker_damage(tracker), n=tracker.n }

  if(not self.enroute_trackers[target])
    self.enroute_trackers[target] = { tracker_data }
  else
    table.insert(self.enroute_trackers[target], tracker_data)
  end
end

--register a friendly tracker launch
function CombatAIMethods:register_tracker_launch(launcher, target)
  local tracker_data = { damage_per=average_weapon_damage(launcher), n=useable_weapons(launcher) }

  if(not self.enroute_trackers[target])
    self.enroute_trackers[target] = { tracker_data }
  else
    table.insert(self.enroute_trackers[target], tracker_data)
  end
end

--register that a weapon is reserved for a potential unlaunched or autofire target
function CombatAIMethods:register_reserved_weapon(reserved_weapon, object)
  local reserved_data = {}
  reserved_data.damage_per = average_weapon_damage(reserved_weapon)
  reserved_data.n = useable_weapons(reserved_weapon)

  if(not self.reserved_weapons[object])
    self.reserved_weapons[object] = { reserved_data }
  else
    table.insert(self.reserved_weapons[object], reserved_data)
  end

  --print("reserving", reserved_weapon.object.name, reserved_weapon.part, "for", launcher.object.name, launcher.part, "damage_per", reserved_data.damage_per, "n", reserved_data.n, "#", #self.reserved_weapons[launcher])
end

function CombatAIMethods:update_ship_priority(target)
  self.ship_priority[target] = self:base_ship_priority(target)
end

---------- Priority Calculation ----------

------ Ships

-- targeting priority for ships is split into a base priority and an adjusted priority
-- the base priority is the combat value of a ship by itself, independent of whatever might be targeting it
-- it is used to determine both the value of a ship as a target and as an asset we might want to defend 
-- from incoming trackers. 
-- In both cases we consider the offensive power of the ship vs it's vulnerability to damage
function CombatAIMethods:base_ship_priority(ship)
  local attack_value = calc_attack_value(ship) + self:boarding_attack_value(ship)
  return attack_value/(1 + calc_defence_value(ship))
end

-- extra attack value granted to a ship based on it's boarding capabilities
function CombatAIMethods:boarding_attack_value(ship)
  local is_boarder, sys = is_combat_boarder(ship)
  if is_boarder
    local boarding_hazard = self:is_hostile(ship) and self.boarding_hazard or self.boarding_power
    local attack_value = boarding_hazard * estimated_boarding_crew_damage(ship)

    --print("ESTIMATED CREW DAMAGE", ship.name, estimated_boarding_crew_damage(ship))

    --if sys == 'Assault Shuttles'
    --  attack_value *= 0.85 --PD fudge factor
    --end

    return attack_value
  end

  return 0
end

-- the adjusted ship priority is the priority of the a ship as a target for the given weapon
-- it is obtained from the base ship priority adjusted for range & distance, attenuation, etc
function CombatAIMethods:get_adjusted_ship_priority(target, weapon)
  local priority = self:base_ship_priority(target)

  local hit_chance = weapon:hit_chance(target) or 100
  priority *= (hit_chance/100)

  local to_target in range_fun
  local dist = to_target(weapon.object, target)

  if(weapon.is_beam)
    local range = weapon:range(dist)
    priority *= beam_attenuate(range)
  end

  if(weapon.is_launcher)
    local launcher_range = weapon:launcher_range()

    --bonus priority if we can deal damage immediately
    if(dist < launcher_range)
      if weapon.is_fighter
        priority *= 3.0 --especially important for fighters, since they are not disposable
      else
        priority *= 1.5
      end
    elseif(target.is_planet)
      priority *= 0.5
    end

    -- unlike missiles which can be relaunched every turn, having fightercraft in flight for more than one turn means lost attacks
    if weapon.is_fighter
      local turns_to_target = dist/launcher_range
      --print(ship.name, weapon.part, target.name, "TTT", turns_to_target)

      if turns_to_target >= 3
        priority = 0
      else
        priority *= 1/max(turns_to_target, 0.5) --at 0.5 we can do a full round trip. closer than that doesn't matter
      end
    end
  end

  -- don't overcommit too many missiles/bombers to a target
  local incoming_trackers = self.enroute_trackers[target]
  if incoming_trackers
    local tracker_damage = 0
    for i, tracker_data in ipairs(incoming_trackers) do
      tracker_damage += tracker_data.damage_per * tracker_data.n
    end

    --TODO factor in the amount of hostile PD
    local structure = max(target.structure - target.structure_damage, 0) --use structure as a proxy for systems hp to avoid extra calculation
    local damage_saturation = ( calc_defence_value(target) + structure) * 1.2
    priority *= clamp(2 - tracker_damage/damage_saturation, 0, 1)
  end

  return priority
end

----- Trackers

-- some trackers have inherent value as things we might want to destroy
-- for example, destroying bombers is as good as destroying ship weapons, so it has value like a ship might
-- missiles OTOH have no value other than to prevent a friendly ship from taking damage, destroying them
-- doesn't prevent our oppoent from launching more
-- NOTE this function is PER TRACKER in the group
local function tracker_inherent_priority_each(tracker)
  if tracker.tracker_type == 'fighter'
    --again, it's important that this is damage/defence akin to ship priorities so that they remain comparable
    return average_tracker_damage(tracker)/tracker.health_per
  end

  return 0
end

--the value of intercepting trackers to prevent damage is based on the value of whatever they're targeting
local function tracker_intercept_priority(tracker, ship_priority_table)
  -- get the priority of the ship this tracker is attacking
  if is_ship(tracker.dest) and tracker.dest ~= tracker.src
    return ship_priority_table[tracker.dest]
  end

  -- in case we want to protect our own trackers from intercept
  if is_tracking_object(tracker.dest) and tracker.dest.empire == current_side()
    local friendly_tracker = tracker.dest
    local inherent_priority = tracker_inherent_priority_each(friendly_tracker)*friendly_tracker.n
    return max(inherent_priority, tracker_intercept_priority(friendly_tracker, ship_priority_table))
  end

  return 0
end

-- gets the priority of a tracker as a target for the given weapon
function CombatAIMethods:get_tracker_priority(tracker, weapon)

  -- ignore trackers that are already targeted by another tracker
  -- important to prevent every interceptor being launched against a single bomber group when there are many
  local already_targeted = 0
  local incoming_trackers = self.enroute_trackers[tracker]
  if incoming_trackers
    for i, tracker_data in ipairs(incoming_trackers) do
      -- account for the fact that some trackers may take more than one hit to destroy
      -- while assuming that one hit can destroy at most one tracker
      already_targeted += min(tracker_data.damage_per/tracker.health_per, 1) * tracker_data.n
    end
  end
  already_targeted = floor(already_targeted)

  local tracker_n = max(tracker.n - already_targeted, 0)

  -- inherent priority
  local inherent_priority = tracker_inherent_priority_each(tracker) * tracker_n

  -- intercept priority
  local intercept_priority = tracker_intercept_priority(tracker, self.ship_priority)

  -- apply a multipler that considers the amount of damage we are preventing
  -- versus the amount of damage to ships we could do with this weapon
  local weapon_damage = average_weapon_damage(weapon) * useable_weapons(weapon)
  local tracker_damage = average_tracker_damage(tracker) * tracker_n
  intercept_priority *= tracker_damage/weapon_damage

  -- other modifiers
  local priority = max(inherent_priority, intercept_priority)

  local hit_chance = weapon:hit_chance(tracker) or 100
  priority *= (hit_chance/100)

  -- fightercraft in flight cannot do anything else, so factor in distance
  if weapon.is_fighter
    local to_target in range_fun
    local dist = to_target(weapon.object, tracker)
    local launcher_range = weapon:launcher_range()
    local turns_to_target = dist/launcher_range

    if dist < launcher_range
      priority *= 3.0
    end

    if turns_to_target >= 3
      priority = 0
    else
      priority *= 1/max(turns_to_target, 0.5)
    end

    -- if the tracker is a craft, check that we can make it there in time
    -- so that we don't end up chasing it all the way to dock
    if tracker.tracker_type == 'fighter'
      local total_move_time in motion_function
      local tracker_range = tracker.speed*total_move_time
      local turns_to_dest = to_target(tracker, tracker.dest)/tracker_range
      if turns_to_target > ceil(turns_to_dest)
        priority *= 0.5
      end
    end
  end

  return priority
end

----- Launchers

-- Assigning priorities to hostile launchers allows the AI to make use of autofire
-- It is also very important for interceptors, since they may not always be able to "fire"
-- every turn, it can make a big difference to wait until a dangerous tracker is launched 
-- instead of wasting turns attacking some worthless target that is in combat now

-- very much like tracker_inherent_priority_each()
local function launcher_inherent_priority_each(launcher)
  if launcher.info.type == 'fighter'
    return average_weapon_damage(launcher)/launcher.info.health
  end

  return 0
end

-- this is very similar to get_tracker_priority(), except we are considering a tracker not yet launched
function CombatAIMethods:get_launcher_priority(launcher, weapon, intercept_priority)

  --don't want all our interceptors/PD reserved for the same launcher
  --pretty much the same calculation as for trackers
  local already_reserved = 0
  local reserved_weapons = self.reserved_weapons[launcher]
  if reserved_weapons
    for i, reserved_data in ipairs(reserved_weapons) do
      already_reserved += min(reserved_data.damage_per/(launcher.info.health or 6), 1) * reserved_data.n
    end
  end
  already_reserved = floor(already_reserved)

  -- the number of trackers the launcher can launch
  -- we are considering what they might do NEXT turn, so don't use useable_weapons()
  -- it doesn't matter that they fired this turn
  local launcher_n = max(alive_weapons(launcher) - launcher.missing - already_reserved, 0)

  -- inherent priority
  local inherent_priority = launcher_inherent_priority_each(launcher) * launcher_n

  -- intercept priority
  -- as an approximation, use our most valuable ship as the hypothetical target for this launcher
  intercept_priority = intercept_priority or self.highest_friendly_priority

  local weapon_damage = average_weapon_damage(weapon) * useable_weapons(weapon)
  local launcher_damage = average_weapon_damage(launcher) * launcher_n
  intercept_priority *= launcher_damage/weapon_damage

  local priority = max(inherent_priority, intercept_priority)

  if weapon.is_fighter
    --this is a little different than the calculation in get_tracker_priority(), for an important reason: 
    --here we are dealing with a hypothetical tracker that might launch NEXT turn
    --we make the assumption that this tracker will move in our general direction, so we subtract
    --it's launcher_range when determining turns_to_target, thus reducing the range penalty.
    --also don't care about dist < launcher_range
    local to_target in range_fun
    local dist = to_target(weapon.object, launcher.object)
    
    local launcher_range = launcher:launcher_range()
    local fighter_range = weapon:launcher_range()

    local turns_to_target = max(dist - launcher_range, 0)/fighter_range

    if turns_to_target >= 3
      priority = 0
    else
      priority *= 1/max(turns_to_target, 0.5)
    end
  end

  print(weapon.object.name, weapon.part, launcher.object.name, launcher.part, "inherent", inherent_priority, "intercept", intercept_priority)

  return priority
end

-- used when deciding to reserve PD for assault shuttles
function CombatAIMethods:get_assault_shuttle_priority(boarder, weapon)
  local shuttle_health_per = 6 --hardcoded in boarding_actions.lua

  --don't want all our interceptors/PD reserved for the same boarder
  --pretty much the same calculation as for trackers
  local already_reserved = 0
  local reserved_weapons = self.reserved_weapons[boarder]
  if reserved_weapons
    for i, reserved_data in ipairs(reserved_weapons) do
      already_reserved += min(reserved_data.damage_per/shuttle_health_per, 1) * reserved_data.n * 0.5
    end
  end
  already_reserved = floor(already_reserved)

  --le sigh
  local base_n=0
  for s in systems(boarder)
    if s.parent_type=='Assault Shuttles'
      base_n+=3
    end
  end

  return self:boarding_attack_value(boarder) * max(base_n - already_reserved, 0)/base_n
end

---------- END Priority Calculation ----------


-- for debugging
local function print_target_table(ship, weapon, target_table, n)
  print(string.format("TARGETS TABLE for %s, %s", ship.name, weapon.part))
  for i, data in ipairs(target_table) do
    local name
    if data.target.is_weapon
      name = string.format("%s.%s", data.target.object.name, data.target.part)
    else
      name = data.target.name
    end
    print(string.format("%s : %.3f %s", name, data.priority, data.reserve and "(RESERVE)" or ""))
    if(n and i >= n) break end
  end
end

local function estimated_ship_move(ship)
  if ship.cant_move
    return 0
  end

  --as a fudge factor, assume a 45 degree turn
  local turn_cost
  if ship.s < 4
    turn_cost = 25/4
  elseif ship.s == 4
    turn_cost = 25/4*0.7
  else
    turn_cost = 25/4/3
  end
  return max(ship.max_fuel - turn_cost, 0)
end

function CombatAIMethods:attack_as_able(ship)
  local weapons = ship.weapons
  if not weapons
    return false
  end

  local fire_map={}

  -- populate fire_map
  local j=1
  while (weapons[j]) do
    local w=weapons[j]
    j=j+1

    local n = w:potential_shots()

    if(n<=0) goto continue end

    -- compile a table of targets by priority
    local target_table = {}


    if w.is_interceptor

      -- consider reserving interceptors for trackers not yet launched
      for i, launcher in ipairs(self.hostile_launchers) do
        target_table[#target_table+1] = { reserve=true, target=launcher.object, priority=self:get_launcher_priority(launcher, w) }
      end

    elseif w.is_pd

      -- consider reserving PD for trackers not yet launched, but only if they are within launch distance of us
      local to_target in range_fun
      for i, launcher in ipairs(self.hostile_launchers) do
        local dist = to_target(launcher.object, ship)
        local launcher_range = launcher:launcher_range() + estimated_ship_move(launcher.object)
        if dist < launcher_range
          --since we are only considering local defence here, use our own priority instead of the highest friendly priority
          local local_defence_priority = (self.ship_priority[ship] + self.highest_friendly_priority)/2
          target_table[#target_table+1] = { reserve=true, target=launcher, priority=self:get_launcher_priority(launcher, w, local_defence_priority) }
        end
      end

      -- consider reserving PD for assault shuttles
      local to_target in range_fun
      for i, boarder in ipairs(self.hostile_shuttles) do
        local dist = to_target(boarder, ship)
        local boarding_range = boarder:boarding_range() + estimated_ship_move(boarder)
        if(0 < dist and dist < boarding_range)
          target_table[#target_table+1] = { reserve=true, target=boarder, priority=self:get_assault_shuttle_priority(boarder, w) }
        end
      end

    end

    -- consider targeting trackers
    if(not w.cant_attack_trackers)
      for i, tracker in ipairs(self.hostile_trackers) do
        if not tracker.dead and w:can_attack(tracker)
          target_table[#target_table+1] = { target=tracker, priority=self:get_tracker_priority(tracker, w) }
        end
      end
    end

    -- consider targeting ships
    for i, target in ipairs(self.targets) do
      if w:can_attack(target)
        target_table[#target_table+1] = { target=target, priority=self:get_adjusted_ship_priority(target, w) }
      end
    end

    if not next(target_table)
      goto continue
    end

    table.sort(target_table,|) function(a,b)
      -- doing something takes precedence over doing nothing, in the case of ties
      -- prevents possible stalemate when all non-PD weapons have been destroyed
      if a.priority == b.priority
        return (not a.reserve) and b.reserve
      end
      return a.priority > b.priority
    end

    if w.is_pd
      print_target_table(ship, w, target_table)
    end

    local target, reserve in target_table[1]
    if reserve
      -- since fighters don't autofire, if we are reserving and there are 
      -- any targets in roundtrip range, attack those instead
      if w.is_fighter
        local to_target in range_fun
        local roundtrip_range = w:launcher_range()*0.5
        for i, item in ipairs(target_table) do
          if(not item.reserve and to_target(ship, item.target) < roundtrip_range)
            local target in item
            fire_map[#fire_map+1] = {..w,..n,..target}
            break
          end
        end
      end

      -- reserve this weapon
      self:register_reserved_weapon(w, target)
      goto continue 
    end

    fire_map[#fire_map+1] = {..w,..n,..target}

    ::continue::
  end


  -- fire weapons in batches based on whether or not they're
  -- weapon_group_equiv.
  if fire_map[1]

    if ship.cloaked
      action.decloak {ship}
    end

    table.sort(fire_map,|) function(a,b)
        return weapon_group_order(a.w,b.w)
    end

    j=1
    for i=1,100 
      if not fire_map[j] break end
      local w in fire_map[j]      
      local push, sub_fire = pusher {}

      while fire_map[j] and weapon_group_equiv(w,fire_map[j].w)
        push | fire_map[j]
        j++
      end

      event_sep_scope(.35,action.fire_weapons,unpack(sub_fire))

      -- after attacking, update the target's priority to reflect possible battle damage
      local w, target in sub_fire[1]
      if is_ship(target)
        if target.dead
          self:drop_target(target)
        else
          self:update_ship_priority(target)
        end
      end
      -- also register the launch of any new trackers
      if w.is_launcher
        self:register_tracker_launch(w, target)
      end
    end
    
  end


  return #fire_map>0
end

-- choose the best of the three closest targets
function CombatAIMethods:choose_move_target(ship)
  local iter_targets = function()
    local i = 0
    return function(s,v)
      i += 1
      if(i <= #self.targets)
        return self.targets[i]
      end
    end
  end

  local closest = closest_objects(iter_targets, object_center(ship))

  local best_target
  local best_priority
  for i=1,min(3,#closest) do
    local target = closest[i]
    local priority = self.ship_priority[target]
    if(not best_target or (priority > best_priority))
      best_target = target
      best_priority = priority
    end
  end

  return best_target
end

function CombatAIMethods:close_and_attack(ship)

  local move_target

  -- if our move target dies, choose a new one and repeat
  while(not ship.dead and (not move_target or move_target.dead)) do

    if(not move_target or move_target.dead or move_target.defenseless or not enemies(ship,move_target) )
      move_target = self:choose_move_target(ship) or nearest(current_enemy_ships, object_center(ship)) 
    end

    -- if there are no valid targets left, exit the loop
    if(not move_target)
      break
    end

    --print("stand_off_and_bombard", stand_off_and_bombard, ship.name)
    if ship.fuel > 0 and not ship.cant_move and not stand_off_and_bombard

      -- move and attack the current target
      local attack_dist = desired_attack_dist(ship)
      if(not attack_dist)
        return
      end

      -- if standing off, continue to move ships, but keep a greater distance
      -- this will allow PD ships and pickets to move ahead of carriers and missile ships
      --[[
      if stand_off_and_bombard
        attack_dist = attack_dist*2 + (desired_attack_dist(move_target) or 8)
        if move_target.is_planet
          attack_dist *= 3
        end
      end
      --]]

      local x,y,r,r2 = find_min_facing_move_L2(ship,move_target,attack_dist)

      if(x)
        if length_to(ship, move_target) <= attack_dist
          --print(ship.name, "NOT MOVING, WITHIN ATTACK DISTANCE", length_to(ship, move_target), "<=", attack_dist)
          x,y = ship.x, ship.y --hack to ensure ships turn in place without moving, causes them to wiggle but that's better than what it was doing before
        end

        action.move_ships{ship, x, y, r}

        if(ship.dead) break end

        if(r2 and r2~=r and ship_can_rotate(ship,r2))
          if(ship.speed==0)
            error("that's very strange")
          end

          action.move_ships{ship, x, y, r2}
        end
      end
    end

    if(ship.dead) break end

    -- combat boarders attempt boarding before firing weapons
    -- and take more effort in prioritizing their boarding target
    if(not ship.has_boarded and is_combat_boarder(ship))
      local boarding_targets = { { target=move_target, priority=boarding_priority(move_target) } }
      for i, target in ipairs(self.targets) do
        if target ~= move_target and can_board(ship,target)
          boarding_targets[#boarding_targets+1] = { target=target, priority=boarding_priority(target) }
        end
      end

      table.sort(boarding_targets,|) function(a,b)
        return a.priority > b.priority
      end

      --[[
      print(string.format("BOARDING TARGETS for %s", ship.name))
      for i, item in ipairs(boarding_targets) do
        print(string.format("%s : %.3f", item.target.name, item.priority))
      end
      --]]

      for i,item in ipairs(boarding_targets) do
        if(not self:capture_if_able(ship, item.target)) break end
      end
    end

    -- actually do the attacking
    local attacks = self:attack_as_able(ship)

    -- set a new target, if this one has been destroyed
    if(move_target.dead)
      self:drop_target(move_target)
      goto continue
    end

    if not ship.cant_electrocyte
      try_use_electrocyte(ship,move_target)
    end

    -- in this case, we should probably try to board the target,
    -- or, if we can't board the target, board the closet ship.
    if(not move_target.dead and not ship.has_boarded)

      -- don't attempt to board in the thick of combat, unless we have a boarding module
      if(is_combat_boarder(ship) or (not attacks or all_enemies_are_immobile()))
        if(self:capture_if_able(ship,move_target))
          self:capture_closest(ship)
        end
      end
    end

    :: continue::
  end

end

-- calls capture_if_able, then drops the target if it was captured
-- otherwise ships might continue to fire on the newly captured target
function CombatAIMethods:capture_if_able(ship, target)
  local rval = capture_if_able(ship, target)
  if(not enemies(ship, target))
    self:drop_target(target)
  end
  return rval
end

-- local we have to duplicate
function CombatAIMethods:capture_closest(ship)
  if(ship.dead) return end

  local targets=closest_objects(current_enemy_ships,object_center(ship))
  for i,target in ipairs(targets) do
    if(ship.dead) return end

    if(not target.dead and not self:capture_if_able(ship,target))
      break
    end
  end
end

local function sorted_current_sides_ships()
  local sort_ships = {}
  for ship in current_sides_ships() do
    sort_ships[#sort_ships+1] = { ship=ship, move_order=calc_attack_value(ship) }
  end
  table.sort(sort_ships,|) function (a,b)
    return a.move_order > b.move_order
  end

  local ships = {}
  for i=1,#sort_ships do
    ships[#ships+1] = sort_ships[i].ship
  end

  return ships
end

local function inside_do_turn()
  if battle_is_over()
    return
  end

  -- local ships=closest_objects(current_sides_ships, enemy_center())
  local ships = sorted_current_sides_ships()

  local try_retreat_all = get_ai_hint() =='retreat_all'
  --print(..stand_off_and_bombard)
  if not try_retreat_all
    --limit bombardment to the first 30 turns
    if all_enemies_are_immobile() and bombardment_available() and current_side_nturns() <= 20
      print('standing off')
      stand_off_and_bombard = true
    else
      stand_off_and_bombard = false
    end
  end

  local ai_turn = CombatAIMethods.init_turn()

  for i,ship in ipairs(ships) do
    if( battle_is_over() )
      sprint("looks like the battle is over")
      break
    end

    if(ship.dead)
      goto continue
    end
    if wants_to_retreat(ship, try_retreat_all)
      action.retreat(ship)
    else
      ai_turn:close_and_attack(ship)
    end

    ::continue::
  end

  if try_retreat_all or (last_event_command()=="end_turn(  )"  and not stand_off_and_bombard) or current_side_nturns() > 40

    console("should probably offer a cease fire", last_event_command(), current_side_nturns() )
    local empire = current_side()
    local attack = all_attacks[get_attack_id()]

    for other_empire in pairs(attack.sides) 
      if Battle.enemy(empire,other_empire) and not Battle.get_cease_fire_offer(empire,other_empire)
        console("offering a cease fire to ",other_empire,empire)
        action.offer_cease_fire(empire,other_empire)
      end

      if empire~=other_empire
        print('cease1', empire, other_empire, objects.cease_fire_offers[ 
          empire._ikey ..'_'.. other_empire._ikey ],
          Battle.get_cease_fire_offer(empire,other_empire)
        )

        print('cease2', other_empire.name, empire.name, objects.cease_fire_offers[ 
          other_empire._ikey ..'_'.. empire._ikey ],
          Battle.get_cease_fire_offer(other_empire,empire)
        )
      end
    end
  end
end


function close_and_attack_turn()
  --console("ai turn for", current_side().name, current_side(), current_side_nturns())

  --print('starting', 'temp AI State', v2s | check_var("temp AI State"))
--  console(last_event_command())

  local ok, err = pcall(inside_do_turn)
  if (not ok) and err
    add_exception(tostring | err)
  end

  action.end_turn()

end

-- BUGFIX for ships retreating because they fired all their weapons while reacting to moving trackers
function wants_to_retreat(ship, try_retreat_all)

  if not can_retreat(ship)
    return
  end

  if try_retreat_all
    return true
  end

  -- pirates and marauder's keep their unarmed ships in the
  -- fight.
  --
  -- pirates, because they have no where to go, and marauders,
  -- because they want to keep their troops around in a raid.
  --
  if ship.empire?.monster_faction
    return
  end

  for w in weapons(ship)
   if alive_weapons(w) > 0
    return false
   end

  end

  return true

end
